#!/usr/bin/env python
# coding: utf-8

# # custom_plots
# This notebook collects all the functions required to plot the graphs needed by the other notebooks.
# 
# Author: Luca Urbinati, 02-10-2019
# 
# Sources:
# - Chapter 5: Examples, pag 643, scikit-learn user guide, Release 0.20.3, Mar 01 2019

# ## Import libraries and custom functions

# In[9]:


from libraries import *
from custom_functions import *


# ## Plot settings

# In[10]:


figsize = (8, 6)
#print("figsize: %s" % (figsize,))
dpi = 110
#print("dpi: %d" % dpi)
markersize=6
alpha_fill = 0.2
alpha_overlap = 0.6

all_colors = ['r', 'g', 'c', 'y', 'b', 'm', 'k', 'deepskyblue', 'brown', 'lime', 'peru', 'moccasin', 'blueviolet', 'orange', 'hotpink', 'darkblue', 'aquamarine', 'olive', 'slategray',  'fuchsia', 'greenyellow', 'crimson', 'cornflowerblue', 'papayawhip']
all_markers = ['^', 'o', 'x', 's', '|', 'd', 'p', 'h', '*', '>', '1', 'v', '2', '<', '3', '_', '4', ',', '8', 'D', 'X', 'P', 'H', '.']

SMALL_SIZE = 12
MEDIUM_SIZE = 16
LARGE_SIZE = 18

plt.rc('font', size=MEDIUM_SIZE)         # controls default text sizes
plt.rc('axes', titlesize=MEDIUM_SIZE)    # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=LARGE_SIZE)   # fontsize of the figure title
plt.rc('lines', lw=2)

title_width = 60 # for the subtitle

loc = "best"


# In[11]:


def dummy():
    print("dummy")


# ## Plot functions

# In[12]:


# PLOT LEARNING CURVE (pag 1248) # TESTATO
# To see if the estimator suffers more from a bias or a variance error, if it has a bias problem or a variance problem.

# There are several things you can do:
# Get more data
# Try a smaller set of features (reduce model complexity)
# Try adding/creating more features (increase model complexity)
# Try increasing the regularisation parameter C (increase model complexity)
# Try decreasing the regularisation parameter C (decrease model complexity)

# Determines cross-validated training and test scores for different training set sizes.
# A cross-validation generator splits the whole dataset k times in training and test data. 
# Subsets of the training set with varying sizes will be used to train the estimator and 
# a score for each training subset size and the test set will be computed. 
# Afterwards, the scores will be averaged over all k runs for each training subset size.
# C'è un riassunto su carta che spiega con un disegno.

# train_sizes = relative or absolute numbers of training examples that will be used to generate the learning curve.
# If the dtype is float, it is regarded as a fraction of the maximum size of the training set 
# (that is determined by the selected validation method), i.e. it has to be within (0, 1]. 
# Otherwise it is interpreted as absolute sizes of the training sets. 
# Note that for classification the number of samples usually have to be big enough to contain at least one sample from each class. 
# (default: np.linspace(0.1, 1.0, 5))


def plot_learning_curve(clf, X, y, scorer, scorer_name, cv, title, ylim, n_jobs=1, train_sizes=np.linspace(0.1, 1.0, 10), sf=dummy, save=False):
    """"
    This function plots the learning curve. 
    X, y = development vectors
    clf = use a classifier after hyper-parameter optimization with hyper-parameters set. 
        grid_search_clf is not accepted. Use grid_search_clf.best_estimator_ or clf with best parameters set.
    """
    
    train_sizes, train_scores, test_scores = learning_curve(clf, X, y, cv=cv, n_jobs=n_jobs, train_sizes=train_sizes, scoring=scorer)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)

    plt.figure(figsize=figsize, dpi=dpi)    
    plt.fill_between(train_sizes, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std, alpha=alpha_fill, color="r")
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std, test_scores_mean + test_scores_std, alpha=alpha_fill, color="g")
    plt.plot(train_sizes, train_scores_mean, 'o-', color="r", label="Training score", linestyle='--')
    plt.plot(train_sizes, test_scores_mean, 'o-', color="g", label="Cross-validation score", linestyle='-')
    
    plt.title("Learning Curves \n (%s)" % "\n".join(wrap(title, width=title_width)))
    plt.xlabel("Training set size")
    plt.ylabel(scorer_name)
    plt.ylim(ylim)
    plt.legend(loc=loc);
    plt.grid()
    plt.tight_layout()   
    sf(plt, file_name='learning curve %s %s' % (title, scorer_name), save=save)
    plt.show()


# In[13]:


# PLOT VALIDATION CURVE (pag 1212) # TESTATO

# Determine training and test scores for varying parameter values.
# Plot scores for an estimator with different values of a specified parameter.  
# It's a utility for plotting the results of gridSearchCV.

def plot_validation_curve_data(data, labels, param_name, param_range, scorer_name, start, stop, title, ylim, n_folds, sf, save=False):
    """
    This function plots the validation curve of one scorer, also a custom one. 
    grid_search_clf = classifier obtained after gridSearchCV()
    """
    #train_scores, valid_scores = validation_curve(clf, X, y, param_name=param_name, param_range=param_range, cv=cv, scoring=scorer, n_jobs=1)

    #results = grid_search_clf.cv_results_
    
    plt.figure(figsize=figsize, dpi=dpi)
    
    for d, label, color, marker in zip(data[start:stop+1], labels[start:stop+1], all_colors[start:stop+1], all_markers[start:stop+1]):
        #train_scores_mean = results['mean_train_score']
        #train_scores_std = results['std_train_score']
        valid_scores_mean = d['mean_test_score']
        valid_scores_std = d['std_test_score']
        
        #loss = results['param_loss']
        #penalty = results['param_penalty']
        #label = "%s, %s" % (loss, penalty)
        #plt.semilogx(param_range, train_scores_mean, label="Training score", color="darkorange")
        #plt.fill_between(param_range, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std, alpha=alpha_fill, color="darkorange")
        plt.semilogx(param_range, valid_scores_mean, label="Cross-Validation Score %s" % label, color=color, marker=marker, alpha=alpha_overlap)
        #plt.fill_between(param_range, valid_scores_mean - valid_scores_std, valid_scores_mean + valid_scores_std, alpha=alpha_fill, color="navy")

        best_score = np.max(valid_scores_mean)

        for mean, std, params in zip(valid_scores_mean, valid_scores_std, param_range):
            if(mean==best_score):
                print("%s = %s (± %f) for %s, %s=%s (color=%s)" % (scorer_name, mean, std, label, param_name, params, color))
                #plt.plot(params, mean, 'x', c='r', markersize=markersize)   

    plt.title("Validation Curves vs. %s \n (%s)" % (param_name, "\n".join(wrap(title, width=title_width))))
    plt.xlabel(param_name)
    plt.ylabel(scorer_name)
    plt.ylim(ylim)
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='validation curves vs %s (%s) %s' % (param_name, title, scorer_name), save=save)
    plt.show()


# In[2]:


# PLOT VALIDATION CURVE (pag 1212) # TESTATO

# Determine training and test scores for varying parameter values.
# Plot scores for an estimator with different values of a specified parameter.  
# It's a utility for plotting the results of gridSearchCV.

def plot_validation_curve_results(results, param_name, param_range, scorer_name, title, ylim, log_scale, n_folds, sf, save=False, nested_mean=None, best_param=None):
    """
    This function plots the validation curve of one scorer, also a custom one. 
    grid_search_clf = classifier obtained after gridSearchCV()
    """
    #train_scores, valid_scores = validation_curve(clf, X, y, param_name=param_name, param_range=param_range, cv=cv, scoring=scorer, n_jobs=1)

    #results = grid_search_clf.cv_results_

    train_scores_mean = results['mean_train_score']
    train_scores_std = results['std_train_score']
    valid_scores_mean = results['mean_test_score']
    valid_scores_std = results['std_test_score']

    plt.figure(figsize=figsize, dpi=dpi)
    
    if(log_scale==True):
        plt.semilogx(param_range, train_scores_mean, label="Training Score", color="darkorange", linestyle='--')
        plt.fill_between(param_range, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std, alpha=alpha_fill, color="darkorange")
        plt.semilogx(param_range, valid_scores_mean, label="Cross-Validation Score", color="navy", linestyle='-')
        plt.fill_between(param_range, valid_scores_mean - valid_scores_std, valid_scores_mean + valid_scores_std, alpha=alpha_fill, color="navy")
    else:
        plt.plot(param_range, train_scores_mean, label="Training Score", color="darkorange", linestyle='--')
        plt.fill_between(param_range, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std, alpha=alpha_fill, color="darkorange")
        plt.plot(param_range, valid_scores_mean, label="Cross-Validation Score", color="navy", linestyle='-')
        plt.fill_between(param_range, valid_scores_mean - valid_scores_std, valid_scores_mean + valid_scores_std, alpha=alpha_fill, color="navy")
    
    best_score = np.max(valid_scores_mean)

    for mean, std, params in zip(valid_scores_mean, valid_scores_std, param_range):
        if(mean==best_score):
            print("%s (± %f) for %r" % (mean, std, params))
            plt.plot(params, mean, 'x', c='r', markersize=markersize)   
        
    if(nested_mean != None):
        #plt.axhline(y=nested_mean, color='r', linestyle=':', label="Nested CV score for C=%s" % best_C)
        plt.plot(best_param, nested_mean, 'o', c='r', markersize=markersize, label="Nested Cross-Validation Score for %s=%s" % (param_name, best_param))

    plt.title("Validation Curves vs. %s \n (%s)" % (param_name, "\n".join(wrap(title, width=title_width))))
    plt.xlabel(param_name)
    plt.ylabel(scorer_name)
    plt.ylim(ylim)
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='validation curve vs %s (%s) %s' % (param_name, title, scorer_name), save=save)
    plt.show()


# In[15]:


# PLOT VALIDATION CURVE FOR MULTIPLE SCORERS (pag 1212, 1228) # TESTATO

def plot_validation_curve_multiple(grid_search_clf, nested_means, param_name, param_range, selected_scorers, color_array, title, ylim, log_scale, n_folds, sf, save=False):
    """
    This function plots the validation curve of one or more library predefined scorers. 
    Each selected_scorers has a color_array for its train and test data.
    grid_search_clf = classifier obtained after gridSearchCV()
    """
    
    results = grid_search_clf.cv_results_

    plt.figure(figsize=figsize, dpi=dpi)

    n_scores = len(selected_scorers)
    
    for i, scorer, color in zip(np.arange(n_scores), selected_scorers, color_array):
        for sample, style in (('train', '--'), ('test', '-')):
             
            sample_score_mean = results['mean_%s_%s' % (sample, scorer)] 
            sample_score_std = results['std_%s_%s' % (sample, scorer)]
            #std_error = sample_score_std / np.sqrt(n_folds)
            plt.fill_between(param_range, sample_score_mean - sample_score_std, sample_score_mean + sample_score_std, alpha=alpha_fill if sample == 'test' else 0, color=color)
            if(log_scale==True):
                plt.semilogx(param_range, sample_score_mean, style, color=color, alpha=1, lw=lw if sample == 'test' else 1, label="%s (%s)" % (scorer, sample))
            else:
                plt.plot(param_range, sample_score_mean, style, color=color, alpha=1, lw=lw if sample == 'test' else 1, label="%s (%s)" % (scorer, sample))
            
            best_score = np.max(results['mean_%s_%s' % (sample, scorer)])
            #best_test_indexes = np.nonzero(results['rank_test_%s' % scorer] == 1)[0]
            
            if(sample=='test'):
                print('scorer: %s' % scorer)
                for mean, std, params in zip(sample_score_mean, sample_score_std, results['params']):
                    if(mean==best_score):
                        print("%s (± %f) for %r" % (mean, std, params))
                        plt.plot(params[param_name], mean, 'x', c='r', markersize=markersize)
                print()

        plt.axhline(y=nested_mean[i], color='k', linestyle=':')
        
    plt.title("Validation Curves \n (%s)" % "\n".join(wrap(title, width=title_width)))
    plt.xlabel(param_name)
    plt.ylabel("Scores")
    plt.ylim(ylim)
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='validation curve multiple %s' % title, save=save)
    plt.show()


# In[27]:


# PLOT CONFUSION MATRIX (pag 1233) # TESTATO

# The diagonal elements represent the number of points for which the predicted label is equal to the true label, 
# while off-diagonal elements are those that are mislabeled by the classifier. 
# The higher the diagonal values of the confusion matrix the better, indicating many correct predictions.

def plot_confusion_matrix(y_true, y_pred, classes, title, sf, normalize=False, plot=True, save=False):
    """ 
    This function prints the confusion matrix and plots it if plot=True. 
    Normalization can be applied by setting normalize=True. 
    """

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    # Only use the labels that appear in the data 
    classes = classes[unique_labels(y_true, y_pred)] 
            
    wasted = (cm[0][1]+cm[1][0])/(cm[0][0]+cm[0][1]+cm[1][0]+cm[1][1]) * 100

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis] * 100
        print("Normalized confusion matrix")
        print(cm)
        title_save = 'Normalized confusion matrix'
        title_plt = title_save+"\n(%s)" % "\n".join(wrap(title, width=title_width))
    else:
        print('Confusion matrix, without normalization') 
        print(cm)
        title_save = 'Confusion matrix'
        title_plt = title_save+"\n(%s)" % "\n".join(wrap(title, width=title_width))

    print("Wasted goods: %0.3f%%" % wasted)
    print()
    
    if(plot == True):
        cmap=plt.cm.Blues
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        im = ax.imshow(cm, interpolation='nearest', cmap=cmap) 
        ax.figure.colorbar(im, ax=ax)
        # We want to show all ticks... 
        ax.set(xticks=np.arange(cm.shape[1]), 
               yticks=np.arange(cm.shape[0]),
               # ... and label them with the respective list entries 
               xticklabels=classes, yticklabels=classes, 
               title=title_plt,
               ylabel='True label',
               xlabel='Predicted label')

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        fmt = '.2f' if normalize else 'd'
        thresh = cm.max() / 2.
        for i in range(cm.shape[0]):
            for j in range(cm.shape[1]):
                ax.text(j, i, format(cm[i, j], fmt),
                        ha="center", va="center",
                        color="white" if cm[i, j] > thresh else "black")
        plt.tight_layout()
        sf(plt, file_name='confusion matrix %s' % title_save, save=save)
        plt.show()


# In[28]:


# PLOT VALIDATION CURVE (pag 1212) # TESTATO

# Determine training and test scores for varying parameter values.
# Plot scores for an estimator with different values of a specified parameter.  
# It's a utility for plotting the results of gridSearchCV.

def plot_validation_curve(grid_search_clf, nested_mean, param_name, param_range, scorer_name, title, ylim, n_folds, sf=dummy, save=False):
    """
    This function plots the validation curve of one scorer, also a custom one. 
    grid_search_clf = classifier obtained after gridSearchCV()
    """
    #train_scores, valid_scores = validation_curve(clf, X, y, param_name=param_name, param_range=param_range, cv=cv, scoring=scorer, n_jobs=1)

    results = grid_search_clf.cv_results_

    train_scores_mean = results['mean_train_score']
    train_scores_std = results['std_train_score']
    valid_scores_mean = results['mean_test_score']
    valid_scores_std = results['std_test_score']

    plt.figure(figsize=figsize, dpi=dpi)

    plt.semilogx(param_range, train_scores_mean, label="Training score", color="darkorange", linestyle='--')
    plt.fill_between(param_range, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std, alpha=alpha_fill, color="darkorange")
    plt.semilogx(param_range, valid_scores_mean, label="Cross-validation score", color="navy", linestyle='-')
    plt.fill_between(param_range, valid_scores_mean - valid_scores_std, valid_scores_mean + valid_scores_std, alpha=alpha_fill, color="navy")

    best_score = np.max(valid_scores_mean)

    for mean, std, params in zip(valid_scores_mean, valid_scores_std, param_range):
        if(mean==best_score):
            print("%s (± %f) for %r" % (mean, std, params))
            plt.plot(params, mean, 'x', c='r', markersize=markersize)   
    
    if(nested_mean != None):
        plt.axhline(y=nested_mean, color='r', linestyle=':')

    plt.title("Validation Curve vs. %s \n (%s)" % (param_name, "\n".join(wrap(title, width=title_width))))
    plt.xlabel(param_name)
    plt.ylabel(scorer_name)
    plt.ylim(ylim)
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='validation curve %s (%s) %s' % (param_name, title, scorer_name), save=save)
    plt.show()


# In[19]:


# PLOT ROC CURVE OF A SPECIFIC CLASS (pag 1245) # TESTATO

# ROC curves are typically used in binary classification to study the output of a classifier.
# ROC curves typically feature true positive rate on the Y axis, and false positive rate on the X axis. 
# This means that the top left corner of the plot is the “ideal” point - a false positive rate of zero, 
# and a true positive rate of one. This is not very realistic, but it does mean that 
# a larger area under the curve (AUC) is usually better.

def plot_roc_curve(y_true, y_score, title, t=0.5, plot2=False, sf=dummy, save=False):
    """ 
    This function plots the roc curve. 
    If plot2=True, also the thresholds are plotted.
    """
    # Compute ROC curve and ROC area for each class
    #for i in range(n_classes):
    fpr, tpr, thresholds = roc_curve(y_true, y_score, drop_intermediate=True) 
    # pos_label: the label of the positive class. 
    # When pos_label=None, if y_true is in {-1, 1} or {0, 1}, pos_label is set to 1.
    roc_auc = roc_auc_score(y_true, y_score)

    plt.figure(figsize=figsize, dpi=dpi)
    plt.plot(fpr, tpr, color='darkorange', label='AUC = %0.3f' % roc_auc) 
    plt.plot([0, 1], [0, 1], color='navy', linestyle='--', label='Random Behavior') 
    plt.xticks(np.arange(round(min(fpr)), round(max(fpr))+0.05, 0.1))
        
    # plot the current threshold on the line
    i_max = np.argmin(np.abs(thresholds - t))
    plt.plot(fpr[i_max], tpr[i_max], 'x', c='r', markersize=markersize)
    
    if(plot2 == True): # http://abhay.harpale.net/blog/machine-learning/threshold-tuning-using-roc/
        min_t = np.min(thresholds)
        max_t = np.max(thresholds)
        # create the axis of thresholds (scores)
        ax2 = plt.gca().twinx()
        ax2.plot(fpr, thresholds, markeredgecolor='r',linestyle='dashed', color='r')
        ax2.set_ylabel('Threshold',color='r')
        ax2.set_ylim([thresholds[-1],thresholds[1]])
        #ax2.set_ylim([min_t,max_t])
        ax2.set_xlim([fpr[0],fpr[-1]])
        print(thresholds)

    plt.title('ROC curve \n (%s)' % "\n".join(wrap(title, width=title_width)))
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.legend(loc="lower right")
    plt.tight_layout()
    sf(plt, file_name='roc curve %s' % title, save=save)
    plt.grid()


# In[20]:


def plot_precision_recall_vs_threshold(precisions, recalls, thresholds, t, title, sf, save=False): # TESTATO
    """ 
    This function plots the precision and recall scores against the decision thresholds. 
    """
    
    plt.figure(figsize=figsize, dpi=dpi)
    
    plt.plot(thresholds, precisions[:-1], "c", label="Precision", linestyle='-')
    plt.plot(thresholds, recalls[:-1], "blueviolet", label="Recall", linestyle='--')
    
    min_t = np.min(thresholds)
    max_t = np.max(thresholds)
    step_t = (max_t - min_t)/10
    plt.xticks(np.round(np.arange(min_t-step_t, max_t+step_t, step_t), 2))

    # plot the current threshold on the line
    #i_max = np.argmin(np.abs(thresholds - t))
    #plt.plot(thresholds[i_max], recalls[i_max], 'x', c='r', markersize=markersize)
    plt.axvline(x=t, c='r', linestyle=':', label="Operating Decision Threshold")
    
    plt.title("Precision and Recall Scores vs. Decision Threshold \n (%s)" % "\n".join(wrap(title, width=title_width)))
    plt.ylabel("Scores")
    plt.xlabel("Decision Threshold")
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='precision-recall vs threshold %s' % title, save=save)
    plt.show()


# In[21]:


# PLOT PRECISION-RECALL CURVE (ONLY BINARY, pag 1251) # TESTATO

# The precision-recall curve shows the tradeoff between precision and recall for different threshold.
# A high area under the curve represents both high recall and high precision, where high precision relates to 
# a low false positive rate, and high recall relates to a low false negative rate. 

# The precision_recall_curve and roc_curve are useful tools to visualize the sensitivity-specificty tradeoff in the classifier.
# They help inform a data scientist where to set the decision threshold of the model to maximize either sensitivity or specificity. 
# This is called the “operating point” of the model.

def plot_precision_recall_curve(y_true, y_score, class_names, title, sf, t=0.5, xlim=[0, 1.05], ylim=[0, 1.05], plot2=False, plot3=False, save=False):
    """ 
    This function plots the precision-recall curve.
    If plot2=True, plot_precision_recall_vs_threshold() is called.
    The values of precision and recall at the classifier's threshold t are shown by the red marker x.
    """

    precisions, recalls, thresholds = precision_recall_curve(y_true, y_score)
    min_t = np.min(thresholds)
    max_t = np.max(thresholds)
    print("min_t: %.3f, max_t: %.3f" % (min_t, max_t))
        
    y_pred_adj = adjusted_prediction(y_score, t)
    
    average_precision = average_precision_score(y_true, y_score)
    print('Average precision-recall score: {0:0.2f}'.format(average_precision))

    print()
    print("--> decision threshold: %0.3f" % t)
    plot_confusion_matrix(y_true, y_pred_adj, classes=class_names, title=title, sf=sf, normalize=False, plot=plot2, save=save)
    plot_confusion_matrix(y_true, y_pred_adj, classes=class_names, title=title, sf=sf, normalize=True, plot=plot2, save=save)
                     
    plt.figure(figsize=figsize, dpi=dpi)
    plt.step(recalls, precisions, color='b', alpha=alpha_overlap, where='post', label="AP=%0.3f" % average_precision)
    step_kwargs = {'step': 'post'}
    plt.fill_between(recalls, precisions, alpha=alpha_fill, color='b', **step_kwargs)
    
    # plot the current threshold on the line
    i_max = np.argmin(np.abs(thresholds - t))
    plt.plot(recalls[i_max], precisions[i_max], 'x', c='r', markersize=markersize)
    
    plt.title('Precision-Recall Curve \n (%s)' % "\n".join(wrap(title, width=title_width)))
    plt.xlabel('Recall')
    plt.xlim(xlim); 
    plt.ylabel('Precision')
    plt.ylim(ylim);
    plt.legend(loc=loc)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='precision-recall curve %s' % title, save=save)
    plt.show()
    
    if(plot3 == True):
        plot_precision_recall_vs_threshold(precisions, recalls, thresholds, t, title, sf, save=save)
    
    return y_pred_adj, thresholds


# In[23]:


# from CV on diabetes dataset example (pag 1035)

def plot_nested_cv_score(nested_means, nested_stds, scorer, scorer_name, title, ylim, n_folds, sf, save=False):
    '''This function plots the nested scores as a function of the random generator seed.'''
    
    valid_scores_mean = scores_mean = nested_means
    valid_scores_std = nested_stds
    x = np.arange(0, len(valid_scores_mean), 1)

    plt.figure(figsize=figsize, dpi=dpi)

    plt.plot(x, valid_scores_mean, color="navy", linestyle='-')
    plt.fill_between(x, valid_scores_mean - valid_scores_std, valid_scores_mean + valid_scores_std, alpha=alpha_fill, color="navy")

    best_score = np.max(valid_scores_mean)

    for mean, std, params in zip(valid_scores_mean, valid_scores_std, x):
        if(mean==best_score):
            half_range = confidence_interval(mean, std, n_samples=n_folds)
            print("%s (± %f) for %r" % (mean, half_range, params))
            plt.plot(params, mean, 'x', c='r', markersize=markersize)   

    plt.xticks(np.arange(min(x), max(x)+0.05, 1))
    plt.title("Nested Cross Validations vs. random_state \n (%s)" % "\n".join(wrap(title, width=title_width)))
    plt.xlabel('random_state')
    plt.ylabel(scorer_name)
    plt.ylim(ylim)
    plt.grid()
    plt.tight_layout()
    sf(plt, file_name='nested cross validation of %s vs random_state' % (scorer_name), save=save)
    plt.show()


# In[24]:


def plot_2D_pca_projection(X, y, class_names, sf, save=False):
    '''This function plots the 2D representation of a dataset (X, y) by using PCA with 2 retained components.'''
    
    y_names = [class_names[0] if target==0 else class_names[1] for target in y]
    target_df = pd.DataFrame({'target': y_names})

    scaler = StandardScaler() # feature_range=(-1,1) or (0, 1)
    scaler.fit(X) # the scaler learns mean and std from the training set
    X_scaled = scaler.transform(X) # perform standardization by centering and scaling: the scaler performs the scaling of the train set using mean and std of the train set

    pca = PCA(n_components=2)
    principalComponents = pca.fit_transform(X_scaled)
    principalDf = pd.DataFrame(data = principalComponents, columns = ['principal component 1', 'principal component 2'])

    finalDf = pd.concat([principalDf, target_df], axis = 1)

    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = plt.axes()
    
    targets = class_names
    colors = ['r', 'g']
    for target, color in zip(targets, colors):
        indicesToKeep = finalDf['target'] == target

        ax.scatter(finalDf.loc[indicesToKeep, 'principal component 1'], 
                   finalDf.loc[indicesToKeep, 'principal component 2'],
                   c = color,
                   s = 50,
                   alpha = alpha_overlap)

    explained_variance = pca.explained_variance_ratio_
    total_explained_variance = sum(explained_variance)*100
    print("Together, the first two principal components contain %0.2f%% of the information" % total_explained_variance)

    ax.set_xlabel('Principal Component 1', fontsize = 15)
    ax.set_ylabel('Principal Component 2', fontsize = 15)
    ax.set_title('2 Components PCA (explained_variance = %0.3f%%)' % total_explained_variance, fontsize=20)
    ax.legend(targets)
    ax.grid()

    file_name = "2 Components PCA"
    plt.tight_layout()
    sf(plt, file_name, save=save)

    plt.show()
    
    return plt


# In[25]:


def plot_3D_pca_projection(X, y, class_names, sf, save=False):
    '''This function plots the 3D representation of a dataset (X, y) by using PCA with 3 retained components.
    The plot is saved from different angle views.'''
   
    y_names = [class_names[0] if target==0 else class_names[1] for target in y]
    target_df = pd.DataFrame({'target': y_names})

    scaler = StandardScaler() # feature_range=(-1,1) or (0, 1)
    scaler.fit(X) # the scaler learns mean and std from the training set
    X_scaled = scaler.transform(X) # perform standardization by centering and scaling: the scaler performs the scaling of the train set using mean and std of the train set

    pca = PCA(n_components=3)
    principalComponents = pca.fit_transform(X_scaled)
    principalDf = pd.DataFrame(data = principalComponents, columns = ['principal component 1', 'principal component 2', 'principal component 3'])

    finalDf = pd.concat([principalDf, target_df], axis = 1)

    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = Axes3D(fig)

    targets = class_names
    colors = ['r', 'g']
    for target, color in zip(targets,colors):
        indicesToKeep = finalDf['target'] == target

        ax.scatter(finalDf.loc[indicesToKeep, 'principal component 1'], 
                   finalDf.loc[indicesToKeep, 'principal component 2'], 
                   finalDf.loc[indicesToKeep, 'principal component 3'],
                   c = color,
                   s = 50,
                   alpha = alpha_overlap)

    explained_variance = pca.explained_variance_ratio_
    total_explained_variance = sum(explained_variance)*100
    print("Together, the first three principal components contain %0.2f%% of the information" % total_explained_variance)

    ax.set_xlabel('Principal Component 1', fontsize = 15)
    ax.set_ylabel('Principal Component 2', fontsize = 15)
    ax.set_zlabel('Principal Component 3', fontsize = 15)
    ax.set_title('3 Components PCA (explained_variance = %0.3f%%)' % total_explained_variance, fontsize=20)
    ax.legend(targets, loc='lower left')
    ax.grid()
    plt.tight_layout()
    
    for ii in range(0,360,40):
        ax.view_init(elev=10, azim=ii)
        fig_number = ii/10
        file_name = "3 Components PCA degree=%d°.png" % fig_number
        sf(plt, file_name, save=save)

    plt.show()


# In[26]:


def plot_history_all(loss_list, val_loss_list, acc_list, val_acc_list, title, ylim_loss, ylim_acc, sf, save=False):
    '''This functions plots the learning curves for a Keras Neural Network model.
    The first learning curve shows the training and validation score (vs. the number of epochs).
    The second learning curve shows the training and validation loss (vs. the number of epochs).
    If this function is called after each fitting, new epochs are appended to the end 
    of the previous learning curves'''
    
    minimum_loss_index = [np.argmin(loss_list)][-1]
    minimum_val_loss_index = [np.argmin(val_loss_list)][-1]
    maximum_acc_index = [np.argmax(acc_list)][-1]
    maximum_val_acc_index = [np.argmax(val_acc_list)][-1]
    
    if len(loss_list) == 0:
        print('Loss is missing in history')
        return 
    
    epochs = range(1,len(loss_list)+1)
    
    plt.figure(figsize=figsize, dpi=dpi)
    plt.plot(epochs, loss_list, 'r', label='Training loss (min=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (loss_list[minimum_loss_index], minimum_loss_index+1, loss_list[-1], len(loss_list)), linestyle='--')
    plt.plot(epochs, val_loss_list, 'g', label='Validation loss (min=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (val_loss_list[minimum_val_loss_index], minimum_val_loss_index+1, val_loss_list[-1], len(val_loss_list)), linestyle='-')

    plt.title('Loss Curves\n%s' % "\n".join(wrap(title, width=title_width)))
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.ylim(ylim_loss)
    plt.legend()
    plt.tight_layout()
    sf(plt, file_name='loss curves %s' % title, save=save)
    
    plt.figure(figsize=figsize, dpi=dpi)
    plt.plot(epochs, acc_list, 'r', label='Training accuracy (max=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (acc_list[maximum_acc_index], maximum_acc_index+1, acc_list[-1], len(acc_list)), linestyle='--')
    plt.plot(epochs, val_acc_list, 'g', label='Validation accuracy (max=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (val_acc_list[maximum_val_acc_index], maximum_val_acc_index+1, val_acc_list[-1], len(val_acc_list)), linestyle='-')

    plt.title('Learning Curves\n%s' % "\n".join(wrap(title, width=title_width)))
    plt.xlabel('Epochs')
    plt.ylabel('Accuracy')
    plt.ylim(ylim_acc)
    plt.legend()
    plt.tight_layout()
    sf(plt, file_name='learning curves %s' % title, save=save)
    plt.show()


# In[ ]:


def plot_history_all2(loss_list, val_loss_list, acc_list, val_acc_list, validation_freq, title, ylim_loss, ylim_acc, sf, save=False):
    '''This functions plots the learning curves for a Keras Neural Network model.
    The first learning curve shows the training and validation score (vs. the number of epochs).
    The second learning curve shows the training and validation loss (vs. the number of epochs).
    If this function is called after each fitting, new epochs are appended to the end 
    of the previous learning curves'''
    
    plt.rc('legend', fontsize=10)    # legend fontsize
    #title_width = 60 # for the subtitle

    loss_list = loss_list[0:len(loss_list):validation_freq]
    acc_list = acc_list[0:len(acc_list):validation_freq]
    
    minimum_loss_index = [np.argmin(loss_list)][-1]
    minimum_val_loss_index = [np.argmin(val_loss_list)][-1]
    maximum_acc_index = [np.argmax(acc_list)][-1]
    maximum_val_acc_index = [np.argmax(val_acc_list)][-1]
    
    if len(loss_list) == 0:
        print('Loss is missing in history')
        return 
    
    epochs = range(1,len(loss_list)+1)
    
    fig, axs = plt.subplots(2, figsize=(8,12), dpi=dpi)
    fig.canvas.draw()
    
    axs[0].plot(epochs, loss_list, 'r', label='Training loss (min=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (loss_list[minimum_loss_index], (minimum_loss_index+1)*validation_freq, loss_list[-1], len(loss_list)*validation_freq), linestyle='--', alpha=alpha_overlap)
    axs[0].plot(epochs, val_loss_list, 'g', label='Validation loss (min=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (val_loss_list[minimum_val_loss_index], (minimum_val_loss_index+1)*validation_freq, val_loss_list[-1], len(val_loss_list)*validation_freq), linestyle='-', alpha=alpha_overlap)

    axs[0].set_title('Loss Curves\n%s' % "\n".join(wrap(title, width=title_width)))
    axs[0].set(xlabel="Epochs", ylabel="Loss")
    axs[0].set_ylim(ylim_loss)
    axs[0].legend(loc="best")
    axs[0].grid()
    plt.tight_layout()
    
    labels = [item.get_text() for item in axs[0].get_xticklabels()]
    labels = [int(i) if i.isdigit() else i for i in labels]
    labels = [i*validation_freq for i in labels]
    axs[0].set_xticklabels(labels)
    
    epochs = range(1,len(acc_list)+1)
    
    axs[1].plot(epochs, acc_list, 'r', label='Training accuracy (max=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (acc_list[maximum_acc_index], (maximum_acc_index+1)*validation_freq, acc_list[-1], len(acc_list)*validation_freq), linestyle='--', alpha=alpha_overlap)
    axs[1].plot(epochs, val_acc_list, 'g', label='Validation accuracy (max=%0.5f @ epoch=%d, last=%0.5f @ epoch=%d)' % (val_acc_list[maximum_val_acc_index], (maximum_val_acc_index+1)*validation_freq, val_acc_list[-1], len(val_acc_list)*validation_freq), linestyle='-', alpha=alpha_overlap)

    axs[1].set_title('Learning Curves\n%s' % "\n".join(wrap(title, width=title_width)))
    axs[1].set(xlabel="Epochs", ylabel="Accuracy")
    axs[1].set_ylim(ylim_acc)
    axs[1].legend(loc="best")
    axs[1].grid()
    plt.tight_layout()
    
    labels = [item.get_text() for item in axs[1].get_xticklabels()]
    labels = [int(i) if i.isdigit() else i for i in labels]
    labels = [i*validation_freq for i in labels]
    axs[1].set_xticklabels(labels)
    
    sf(plt, file_name='loss curves %s' % title, save=save)
    plt.show()


# In[ ]:


print("custom_plots imported")

